"""
PySoleno by Mariusz Wozniak
Soleno was originally developed at Twente University by Gert Mulder, and upgraded in 1999 by Erik Krooshoop
This code is based on Jeroen van Nugteren C++ implementation of original soleno.
The test are comparing to Jeroen results generated in matlab and saved to text file.

"""

import numpy as np


class PySoleno:

    def __init__(self):
        self.Eps = 1e-12  # calculation accuracy
        self.Tmax = 50  # max number of loops (?)

    # ---------- Field ------------

    def _Sol_Rekenb(self, Zmh, Rm, Rp):
        Hulp = np.sqrt(Rp * Rp + Zmh * Zmh)
        Kc = np.sqrt(Rm * Rm + Zmh * Zmh) / Hulp
        if np.any(Kc < self.Eps):
            Kc = self.Eps
        Mj = 1
        Nj = Kc
        Aj = -1
        Bj = 1
        Pj = Rm / Rp
        Cj = Pj + 1
        if np.any(np.abs(Rm / Rp) < self.Eps):
            Pj = 1
        Dj = Cj * np.sign(Pj)
        Pj = np.abs(Pj)
        Tel = 1
        while np.any(np.abs(Mj - Nj) > (self.Eps * Mj)):
            if np.any(Tel > self.Tmax):
                print("Solution does not converge")
            Xmn = Mj * Nj
            Xmp = Xmn / Pj
            D0 = Dj
            Dj = 2 * (Dj + Xmp * Cj)
            Cj = Cj + D0 / Pj
            B0 = Bj
            Bj = 2 * (Bj + Nj * Aj)
            Aj = Aj + B0 / Mj
            Pj = Pj + Xmp
            Mj = Mj + Nj
            Nj = 2 * np.sqrt(Xmn)
            Tel = Tel + 1
        Brh = (Aj + Bj / Mj) / (Mj * Hulp)
        Bzh = (Cj + Dj / Mj) / ((Mj + Pj) * Hulp) * Zmh
        return Brh, Bzh

    def _Sol_CalcB(self, R, Z, Rin, Rout, Zlow, Zhigh, It, Nlayers):
        # Original Soleno Function (used by improved version)
        Brd = 0
        Bzd = 0
        if np.any(Rin != Rout) and np.any(Zlow != Zhigh):
            Factor = It / (Zhigh - Zlow) / Nlayers
            if np.any(Factor != 0):
                Zm1 = Z - Zlow
                Zm2 = Z - Zhigh
                dR = (Rout - Rin) / Nlayers
                dH = dR / 2
                dE = dH
                for tt in range(1, Nlayers + 1):
                    Rs = Rin - dH + dR * tt
                    if np.any(np.abs(R - Rs) >= dE):
                        # extern
                        Rp = Rs + R
                        Rm = Rs - R
                        Br1, Bz1 = self._Sol_Rekenb(Zm1, Rm, Rp)
                        Br2, Bz2 = self._Sol_Rekenb(Zm2, Rm, Rp)
                        Brd = Brd + (Br2 - Br1) * Factor * Rs
                        Bzd = Bzd + (Bz1 - Bz2) * Factor
                    else:
                        # intern
                        Rp = Rs + Rs - dE
                        Rm = dE
                        Weeg = (Rs - R + dE) / dE / 2
                        Br1, Bz1 = self._Sol_Rekenb(Zm1, Rm, Rp)
                        Br2, Bz2 = self._Sol_Rekenb(Zm2, Rm, Rp)
                        Brd = Brd + (Br2 - Br1) * Weeg * Factor * Rs
                        Bzd = Bzd + (Bz1 - Bz2) * Weeg * Factor
                        Weeg = 1 - Weeg
                        Rp = Rs + Rs + dE
                        Rm = -dE
                        Br1, Bz1 = self._Sol_Rekenb(Zm1, Rm, Rp)
                        Br2, Bz2 = self._Sol_Rekenb(Zm2, Rm, Rp)
                        Brd = Brd + (Br2 - Br1) * Weeg * Factor * Rs
                        Bzd = Bzd + (Bz1 - Bz2) * Weeg * Factor
        return Brd * np.pi * 1e-7, Bzd * np.pi * 1e-7

    def _Sol_CalcB_new(self, R2d, Z2d, Rin, Rout, Zlow, Zhigh, I, Nlayers):
        if Nlayers < 1:  # set layers if not set remotely 5 is default value
            Nlayers = 5
        R = np.ravel(R2d)
        Z = np.ravel(Z2d)
        internal_mask = ((R >= Rin) & (R <= Rout))

        def internal(R, Z, Rin, Rout, Zlow, Zhigh, I, Nlayers):
            Brt_l = []
            Bzt_l = []
            for R, Z in zip(list(R), list(Z)):
                A = (R - Rin) / (Rout - Rin)
                NL1 = int(Nlayers * A) + 1
                NL2 = int(Nlayers * (1 - A)) + 1
                # deel1
                Br2, Bz2 = self._Sol_CalcB(R, Z, Rin, R, Zlow, Zhigh, I * A, NL1)
                Brt, Bzt = self._Sol_CalcB(R, Z, Rin, R, Zlow, Zhigh, I * A, NL1 * 2)
                BZtemp = Bzt + (Bzt - Bz2) / 3  # Extrapolatie naar 1/Nl/Nl=0
                BRtemp = Brt + (Brt - Br2) / 3
                # deel2
                Br2, Bz2 = self._Sol_CalcB(R, Z, R, Rout, Zlow, Zhigh, I * (1 - A), NL2)
                Brt, Bzt = self._Sol_CalcB(R, Z, R, Rout, Zlow, Zhigh, I * (1 - A), NL2 * 2)
                Brt_l.append(BRtemp + Brt + (Brt - Br2) / 3)
                Bzt_l.append(BZtemp + Bzt + (Bzt - Bz2) / 3)  # Extrapolatie naar 1/Nl/Nl=0
            return np.array(Brt_l), np.array(Bzt_l)

        def external(R, Z, Rin, Rout, Zlow, Zhigh, I, Nlayers):
            Br2, Bz2 = self._Sol_CalcB(R, Z, Rin, Rout, Zlow, Zhigh, I, Nlayers)
            Brt, Bzt = self._Sol_CalcB(R, Z, Rin, Rout, Zlow, Zhigh, I, Nlayers * 2)
            Bzt = Bzt + (Bzt - Bz2) / 3  # Extrapolatie naar 1/Nl/Nl=0
            Brt = Brt + (Brt - Br2) / 3
            return Brt, Bzt

        Brt = np.zeros_like(R)
        Bzt = np.zeros_like(R)
        Brt[internal_mask], Bzt[internal_mask] = internal(R[internal_mask], Z[internal_mask], Rin, Rout, Zlow, Zhigh, I,
                                                          Nlayers)
        # Brt[internal_mask], Bzt[internal_mask] = external(R[internal_mask], Z[internal_mask], Rin, Rout, Zlow, Zhigh, I,
        #                                                   Nlayers)
        Brt[~internal_mask], Bzt[~internal_mask] = external(R[~internal_mask], Z[~internal_mask], Rin, Rout, Zlow,
                                                            Zhigh, I, Nlayers)
        Brt = np.reshape(Brt, R2d.shape)
        Bzt = np.reshape(Bzt, R2d.shape)
        return Brt, Bzt

    # ---------- Inductance ------------
    def _Sol_CI(self, R1, R2, ZZ):
        # Function for calculating elliptical integrals
        Imax = 25
        Itel = 0
        Ci = 0
        Rm = R1 - R2
        Rp = R1 + R2
        Zkw = ZZ * ZZ
        Rkw = 4 * R1 * R2
        Kkw = (Rm * Rm + Zkw) / (Rp * Rp + Zkw)
        if Kkw < self.Eps:
            Ci = 1 / (6 * np.arctan(1.0))  # f was removed not sure what it did
        else:
            Kc = np.sqrt(Rkw / (Rp * Rp + Zkw))
            Alfa1 = 1
            Beta1 = np.sqrt(Kkw)
            Q1 = np.abs(Rm / Rp)
            R1eqR2 = Q1 <= self.Eps
            A1 = 1 / (3 * Kc)
            B1 = -Kkw * A1
            C1 = Kc * Zkw / Rkw
            D1 = 0
            A1 = A1 - C1
            if R1eqR2:
                A1 = A1 + C1
                B1 = -B1 - B1
                C1 = 0
            while (abs(Alfa1 - Beta1) > Alfa1 * self.Eps and Itel < Imax):  # && replaced with and
                Itel = Itel + 1
                AlfBe1 = Alfa1 * Beta1
                if not R1eqR2:
                    C2 = C1 + D1 / Q1
                    D2 = AlfBe1 * C1 / Q1 + D1
                    D2 = D2 + D2
                    Q2 = Q1 + AlfBe1 / Q1
                    C1 = C2
                    D1 = D2
                    Q1 = Q2
                Alfa2 = Alfa1 + Beta1
                Beta2 = 2 * np.sqrt(AlfBe1)
                A2 = B1 / Alfa1 + A1
                B2 = B1 + Beta1 * A1
                B2 = B2 + B2
                Alfa1 = Alfa2
                Beta1 = Beta2
                A1 = A2
                B1 = B2
            Ci = (A1 + B1 / Alfa1) / (Alfa1 + Alfa1) + (C1 + D1 / Alfa1) / (Alfa1 + Q1)
            if Itel >= Imax:
                print("No Convergence")
        return Ci

    # Main mutual inductance function returns mutual inductance between coil I and J
    def _CalcMutSub(self, SD, I, J, NlI, NlJ):
        S1 = np.abs(SD['Zh'][I] - SD['Zl'][J])
        S2 = np.abs(SD['Zl'][I] - SD['Zl'][J])
        S3 = np.abs(SD['Zl'][I] - SD['Zh'][J])
        S4 = np.abs(SD['Zh'][I] - SD['Zh'][J])
        S3neS1 = (S3 != S1)
        S4neS2 = (S4 != S2)
        RdivNE = (SD['Ri'][I] != SD['Ri'][J]) or (SD['Ro'][I] != SD['Ro'][J]) or (NlI != NlJ)
        Factor = 0.0000008 * np.pi * np.pi * SD['Nw'][I] / NlI / (SD['Zh'][I] - SD['Zl'][I]) * SD['Nw'][J] / NlJ / (
                SD['Zh'][J] - SD['Zl'][J])
        DrI = (SD['Ro'][I] - SD['Ri'][I]) / NlI
        DrJ = (SD['Ro'][J] - SD['Ri'][J]) / NlJ
        Som = 0

        for K in range(1, NlI + 1):
            R1 = SD['Ri'][I] + (K - 0.5) * DrI
            for L in range(1, NlJ + 1):
                R2 = SD['Ri'][J] + (L - 0.5) * DrJ
                if RdivNE or L <= K:
                    C1 = self._Sol_CI(R1, R2, S1)
                    C2 = self._Sol_CI(R1, R2, S2)
                    C3 = C1
                    C4 = C2
                    if (S3neS1):
                        C3 = self._Sol_CI(R1, R2, S3)
                    if (S4neS2):
                        C4 = self._Sol_CI(R1, R2, S4)
                    HH = R1 * R2
                    HH = np.sqrt(HH * HH * HH)
                    if not RdivNE and (L < K):
                        HH = 2 * HH
                    Som = Som + (C1 - C2 + C3 - C4) * HH
        return Som * Factor

    def _CalcMut(self, SD, Ns, N1, CalcE):
        Mut = np.zeros(shape=(Ns, Ns))  # new
        # Mut = [0]*Ns*Ns
        Msom = 0
        if N1 == Ns:
            N2 = 0
        else:
            N2 = N1
        for I in range(0, Ns):
            for J in range(I, Ns):
                if J < N2 or I >= N1:
                    Hulp = 0
                else:
                    if SD['NN'][I] > 0 or SD['NN'][J] > 0 or CalcE:
                        NlI = SD['NL'][I]
                        NlJ = SD['NL'][J]
                        Hulp2 = self._CalcMutSub(SD, I, J, NlI, NlJ)  # opgegeven aantal lagen
                        Hulp = self._CalcMutSub(SD, I, J, NlI * 2, NlJ * 2)  # dubbel aantal lagen
                        Hulp = Hulp + (Hulp - Hulp2) / 3  # extrapolatie naar 1/NLI/NLJ=0
                        Mut[I, J] = Hulp
                        Mut[J, I] = Hulp
                        # Mut[I+J*Ns] = Hulp
                        # Mut[J+I*Ns] = Hulp
                    # Hulp = Mut[I+J*Ns]
                if (N1 == Ns and I != J):
                    Hulp = 2 * Hulp
                # if (CalcE):
                #     Hulp = Hulp * SD[I].Iw * SD[J].Iw
                # Msom = Msom + Hulp
        if N1 == Ns:
            for I in range(0, Ns):
                SD['NN'][I] = 0
        # return Msom
        return Mut

    # ----------- Public methods ----------------

    def calcB(self, rr, zz, Rin, Rout, Zlow, Zhigh, I, Nturn, Nloop):
        """
        Calculate magnetic field components of a solenoid(s) at points
        :param rr: radial grid coordinates where field to be calculated in m
        :param zz: axial grid coordinates where field to be calculated in m
        :param Rin: (list of) inner radius(es) of solenoid(s) in m
        :param Rout: (list of) outer radius(es) of solenoid(s) in m
        :param Zlow: (list of) lower axial position(s) of solenoid(s) in m
        :param Zhigh: (list of) higher axial position(s) of solenoid(s)in m
        :param I: (list of) current(s) of solenoid(s) in A
        :param Nturn: (list of) number of turn(s) of solenoid(s) - integer
        :param Nloop: number of loops in calculations - higher number = higher accuracy, 5 is usually okay
        :return: Br and Bz are field in the radial and axial direction in T
        """
        rr1 = np.ravel(rr)
        pos_mask = rr1 > 0.
        rrp = rr1[pos_mask]
        rrn = rr1[~pos_mask]
        zz1 = np.ravel(zz)
        zzp = zz1[pos_mask]
        zzn = zz1[~pos_mask]
        Itot = [n * i for n, i in zip(Nturn, I)]
        if len(set(map(len, [Rin, Rout, Zlow, Zhigh, Itot, Nloop]))) == 1:  # check if input lists are the same length
            # print("Inputs are the same length")
            Br = np.zeros_like(rr1)
            Bz = np.zeros_like(rr1)
            for ri, ro, zl, zh, it, nl in zip(Rin, Rout, Zlow, Zhigh, Itot, Nloop):
                br, bz = self._Sol_CalcB_new(rrp, zzp, ri, ro, zl, zh, it, nl)
                Br[pos_mask] = Br[pos_mask] + br
                Bz[pos_mask] = Bz[pos_mask] + bz
                if np.min(rr1) < 0:
                    br, bz = self._Sol_CalcB_new(-rrn, zzn, ri, ro, zl, zh, it, nl)
                    Br[~pos_mask] = Br[~pos_mask] - br
                    Bz[~pos_mask] = Bz[~pos_mask] + bz
            Br = np.reshape(Br, rr.shape)
            Bz = np.reshape(Bz, rr.shape)
        else:
            raise SystemExit("Inputs are not the same length!")
        return Br, Bz

    def calcM(self, Rin, Rout, Zlow, Zhigh, I, Nturn, Nloop):  # kept I to make inputs for calcB and calcM the same
        """
        Calculate mutual inductance of solenoid(s)
        :param Rin: (list of) inner radius(es) of solenoid(s) in m
        :param Rout: (list of) outer radius(es) of solenoid(s) in m
        :param Zlow: (list of) lower axial position(s) of solenoid(s) in m
        :param Zhigh: (list of) higher axial position(s) of solenoid(s)in m
        :param I: (list of) current(s) of solenoid(s) - this is not needed for calculations, added for consistency with calcB method
        :param Nturn: (list of) number of turn(s) of solenoid(s) - integer
        :param Nloop: number of loops in calculations - higher number = higher accuracy, 5 is usually okay
        :return: matrix  is in a standard form in H
        """
        Ns = len(Rin)  # number of coils = total size of matrix i.e. for 2 coils Ns = 4
        SD = {'Ri': Rin, 'Ro': Rout, 'Zl': Zlow, 'Zh': Zhigh, 'Nw': Nturn, 'NL': Nloop,
              'NN': [1] * len(Rin)}  # Nw - number of turns, NL - number of layers, NN - not calculated coil
        return self._CalcMut(SD, Ns, Ns, False)



import numpy as np
import pandas as pd



def read_matlab_magnet(test_number):
    file_name = f'test{test_number}.magnet'
    df = pd.read_csv(file_name, delim_whitespace=True)
    design = [list(df[key]) for key in list(df)]
    return design

def read_matlab_field(test_number):
    file_name = f'test{test_number}.field'
    df = pd.read_csv(file_name, delim_whitespace=True)
    first_col = list(df[list(df)[0]])
    z_d = first_col.count(first_col[0])
    r_d = int(len(first_col) / z_d)
    return [np.reshape(np.array(df[label]), (r_d, z_d), order='F') for label in
            ['r(mm)', 'z(mm)', 'Br(T)', 'Bz(T)']]

def read_matlab_inductance(test_number):
    file_name = f'test{test_number}.inductance'
    df = pd.read_csv(file_name, delim_whitespace=True)
    return df.to_numpy()


import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm
from mpl_toolkits.axes_grid1 import make_axes_locatable

import src.pysoleno as pysol
import matlab_files_handler as mfh

def plot_field(test_number, comp):
    """
    Plot magnetic field or coordinates and compare to matlab output saved as text files
    :param comp: 'Br' or 'Bz' or 'Bmod' or "coor"
    :return: plot on screen
    """

    plots = ['matlab', 'python', '(matlab-python)']
    fig, axs = plt.subplots(nrows=1, ncols=len(plots), figsize=(15, 5))
    rr_m, zz_m, Br_m, Bz_m = mfh.read_matlab_field(test_number)
    Bmod_m = np.sqrt(np.square(Br_m) + np.square(Bz_m))
    B_m = {'Br': Br_m, 'Bz': Bz_m, 'Bmod': Bmod_m, "coor": rr_m + zz_m}
    D_m = mfh.read_matlab_magnet(test_number)  # data for magnet from matlab
    s = pysol.PySoleno()
    Br_p, Bz_p = s.calcB(rr_m, zz_m, *D_m)
    Bmod_p = np.sqrt(np.square(Br_p) + np.square(Bz_p))
    B_p = {'Br': Br_p, 'Bz': Bz_p, 'Bmod': Bmod_p, "coor": rr_m + zz_m}
    B_e = {'Br': Br_m - Br_p, 'Bz': Bz_m - Bz_p, 'Bmod': Bmod_m - Bmod_p, "coor": rr_m + zz_m - (rr_m + zz_m)}
    sets = {}
    data = [B_m, B_p, B_e]
    for pl, d in zip(plots, data):
        sets[pl] = [rr_m, zz_m, d[comp]]
    for ax, title in zip(axs, plots):
        im = ax.pcolor(*sets[title], shading='auto', cmap=cm.get_cmap('jet'))
        # im = ax.pcolormesh(*sets[title], shading='auto', cmap=cm.get_cmap('jet'))
        ax.set_aspect(1)
        ax.set_title(f'{title}: {comp} [T]')
        ax.set_xlim(np.min(sets[title][0]), np.max(sets[title][0]))
        ax.set_ylim(np.min(sets[title][1]), np.max(sets[title][1]))
        ax.set_xlabel("r (m)")
        ax.set_ylabel("z (m)")
        cax = make_axes_locatable(ax).append_axes('right', size='5%', pad=0.05)
        fig.colorbar(im, cax=cax, orientation='vertical')
    plt.tight_layout()
    plt.show()

def manual_magnet_output(calc_pos=True):
    """
    This is dummy output, not used by any functions above :return: tuple with magnet definition, can be unpacked
    in calcB function or inside M function if B is set to false
    :param calc_pos: Set to true to return field calculation positions in return tuple
    :return: Tuple that can be unpacked in Soleno calcB and calcM function
    """
    # definition of magnet
    Rin = [0.13, 0.13]
    Rout = [0.15, 0.15]
    Zlow = [-0.1, 0.05]
    Zhigh = [-0.05, 0.1]
    Nturn = [200, 200]
    I = [200, 200]
    Nloop = [5, 5]
    # definition of target points
    Rmin = -0.4
    Rmax = 0.4
    Zmin = -0.4
    Zmax = 0.4
    r_d = 50
    z_d = 50
    rr, zz = np.meshgrid(np.linspace(Rmin, Rmax, r_d), np.linspace(Zmin, Zmax, z_d))
    if calc_pos:
        return rr, zz, Rin, Rout, Zlow, Zhigh, I, Nturn, Nloop
    else:
        return Rin, Rout, Zlow, Zhigh, I, Nturn, Nloop

if __name__ == '__main__':
    test_number = 1
    for comp in ['Br', 'Bz', 'Bmod']:
        plot_field(1, comp)



import unittest
import numpy as np

import src.pysoleno as pysol
from tests import matlab_files_handler as mfh


class Test_pysoleno(unittest.TestCase):

    test_number = 1
    s = pysol.PySoleno()

    def test_B(self):
        """
        Checks if Br and Bz differ by less than specified thresholds
        """
        B_err = 1e-6  # maximum allowed field error (compared to matlab) - just a guess
        D_m = mfh.read_matlab_magnet(self.test_number)  # data for magnet from matlab
        rr_m, zz_m, Br_m, Bz_m = mfh.read_matlab_field(self.test_number)
        Br_p, Bz_p = self.s.calcB(rr_m, zz_m, *D_m)
        Br_e = Br_m - Br_p
        Bz_e = Bz_m - Bz_p
        Br_e_m = np.max(Br_e)
        Bz_e_m = np.max(Bz_e)
        self.assertLessEqual(Br_e_m, B_err)
        self.assertLessEqual(Bz_e_m, B_err)
        if Br_e_m < B_err and Bz_e_m < B_err:
            print(f'OKAY - Field calculation is test passed with error below Br={Br_e_m:.2e} T and Bz= {Bz_e_m:.2e} T')
        else:
            raise Exception("Field calculation calculation error")

    def test_M(self):
        """
        Checks if M differ by less than specified thresholds
        """
        M_err = 1e-6  # maximum allowed inductance error (compared to matlab) - just a guess
        D_m = mfh.read_matlab_magnet(self.test_number)  # data for magnet from matlab
        M_m = mfh.read_matlab_inductance(self.test_number)
        M_p = self.s.calcM(*D_m)
        M_e = M_m - M_p
        M_e_s = np.max(M_e)
        self.assertLessEqual(M_e_s, M_err)
        if M_e_s < M_err:
            print(f"OKAY - Self-mutual inductance calculation test passed with errors below {M_e_s:.2e} H")
        else:
            raise Exception("Self-mutual inductance calculation error")

if __name__ == '__main__':
    unittest.main()




from setuptools import setup
from setuptools import find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name='pysoleno',
    version="0.0.4",
    author="Mariusz Wozniak",
    author_email="mariusz.wozniak@cern.ch",
    description="Solenoid field and inductance matrix calculator.",
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url="https://gitlab.com/mawoznia/PySoleno",
    keywords={'magnetic field', 'coil', 'solenoid', 'inductance'},
    install_requires=["numpy"],
    extras_require={"dev": ["pandas", "matplotlib",],},
    python_requires='>=3.9',
    include_package_data=True,
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.8"],

)
